/* -*- compile-command: "gcc -o kdm2it kdm2it.c -O2" -*-
 *
 *  kdm2it.c - convert a Ken's Digital Music file to an Impulse Tracker Module
 *
 *  Copyright (C) 2005  Jared Stafford (jspenguin)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Technical notes on conversion
 * -----------------------------
 *
 * Notes:
 *
 * Notes are fairly straightforward to convert: The formula is
 * itnote=kdmnote+35. Note off commands are converted to note cut
 * commands.
 *
 * Tracks/Channels:
 *
 * This gets a little tricky. Each KDM track can have more than one
 * note playing at a time. Although IT *can* more than one note playing
 * on a track (with NNA=continue), once a new note is played, we no
 * longer have control over the old ones, so I allocate one or more IT
 * channels to each KDM track that has note data (some tracks are left
 * blank). First, I iterate through the notes and calculate the maximum
 * polyphony for each track. This is not as straightforward as it
 * seems, because some instruments do not repeat, and therefore do not
 * require note off commands. For tracks that do not contain *any* note
 * off commands (usually drum tracks), I set the NNA to continue, and
 * each channel "expires" on the next row (just in case 2 drums are
 * played at once). For tracks that do contain some note off commands,
 * and the instrument does not repeat, the channel expires after enough
 * rows to play the note at C-5.
 *
 * Effects:
 *
 * This is the trickiest part of all. First of all, for simplicity, I
 * ignore panning effects (I can't usually hear the difference). I
 * handle volume and pitch effects in two ways. The first way is with a
 * table of functions which act on the note being played and subsequent
 * rows of the same channel, until the note is turned off or another
 * note is played. These functions will set various commands in the IT
 * note structure. Since only one effect function can be active (if
 * there were two active one would overwrite the other's output), pitch
 * effects override volume effects.
 *
 * The second way I handle effects is with an envelope table. This is
 * used when no suitable IT commands can be used for an effect. It is a
 * last resort method because for each combination of effects that use
 * envelopes, a new instrument has to be created (although it maps to
 * the same sample to save space).
 *
 * Samples:
 *
 * Samples are fairly straightforward to convert. KDM uses unsigned
 * samples while IT uses signed samples, so the data must be
 * converted. For repstart/repleng I use IT loopbeg and loopend
 * instead of sustain. KDM has a finetune value for each sample which
 * is converted to an IT c5speed with the following formula:
 * (8259*((finetune)+748)/748.0))
 *
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#ifdef _WIN32
#ifndef WIN32
#define WIN32
#endif
#endif
#if !defined(inline) && !defined(__GNUC__)
#define inline _inline
#endif

#ifdef WIN32
#define BYTE_ORDER 1234
#define BIG_ENDIAN 4321
#endif

#define NEW(type) (type*)malloc(sizeof(type))
#define NEWARR(type,size) (type*)malloc(sizeof(type)*size)

#define IT_HEADSIG 0x4d504d49
#define IT_INSTSIG 0x49504d49
#define IT_SAMPSIG 0x53504d49


#define SF_16BIT 2
#define SF_LOOP 16
#define SF_SUSLOOP 32
#define SF_LOOP_PING 64
#define SF_SUSLOOP_PING 128

#define EF_ON 1
#define EF_LOOP_ON 2
#define EF_SUSLOOP_ON 4

#define NOTEAT(pat,row,chan) ((pat)->notes[(row)*(pat)->chans+(chan)])

typedef unsigned char byte;

typedef struct {
	 int len,cap;
	 byte* data;
} buffer;

typedef struct {
	 unsigned int time;
	 byte track,freq,vol1,vol2,frqeff,voleff,paneff;
} kdmnote_t;

typedef struct {
	 int ntracks, nnotes;
	 byte* trinst; // Track instrument data
	 byte* trquant; // Track quantifier data
	 byte* trvol1; // Track volume data (left channel)
	 byte* trvol2; // Track volume data (right channel)

	 kdmnote_t* notes;

} kdmfile_t;

typedef struct {
	 char* name;
	 byte* snd;
	 int len;
	 int repstart;
	 int repleng;
	 int repend;
	 int finetune;
} kdmwave_t;

typedef struct {
	 int nwaves;
	 kdmwave_t* waves;
} kdmwaveset_t;

typedef struct {
	 char dosfn[16];
	 char name[26];
	 int gvol,dvol;
	 int flags;
	 int len;
	 int loopbeg,loopend;
	 int sloopbeg,sloopend;
	 int speed;
	 int vibspeed,vibdepth,vibrate,vibtype;
	 int defaultpan;
	 unsigned char* sampdata;
} it_sampinfo;

typedef struct {
	 int flags;
	 int nnodes;
	 int loopbeg,loopend;
	 int sloopbeg,sloopend;
	 int nodepos[25];
	 int nodeval[25];
} it_envelope;

#define ENV_VOL 0
#define ENV_PAN 1
#define ENV_PITCH 2

typedef struct {
	 char dosfn[16];
	 char name[26];
	 int nna;
	 int dct,dca;
	 int fadeout;
	 int pps,ppc;
	 int gvol,dpan;
	 int map_note[120];
	 int map_samp[120];
	 it_envelope env[3];
} it_instinfo;

typedef struct {
	 unsigned char note,inst,volpan,eff,effval;
} note_t;

typedef struct {
	 int rows,chans;
	 note_t* notes;
} it_pattern;

typedef struct {
	 char songname[26];
	 int ordnum, insnum, smpnum, patnum;
	 const char* message;
	 int globvol;
	 int ispeed, itempo;
	 int pansep;
	 unsigned char pantable[64];
	 unsigned char voltable[64];
	 unsigned char* order;
	 it_instinfo** insts;
	 it_sampinfo** samples;
	 it_pattern** pats;
} it_info;



typedef struct {
	 short* buf;
	 unsigned int len;
	 
} wave_t;

void resizebuffer(buffer* buf, int nl) {
	 int ccap=buf->cap;
	 while (nl>ccap) {         
	ccap+=ccap>>2;            
	 }               
	 buf->cap=ccap;
	 buf->data=(byte *)realloc(buf->data,ccap);

}
void trimbuffer(buffer* buf) {
	 if (buf->cap>buf->len) {
	buf->data=(byte *)realloc(buf->data,buf->len);
	buf->cap=buf->len;
	 }
}
inline void checkbuffer(buffer* buf, int nl) {
	 if (nl>buf->cap) resizebuffer(buf,nl);
}
buffer* allocbuf(int cap) {
	 buffer* ret=(buffer *)malloc(sizeof(buffer));
	 ret->cap=cap;
	 ret->len=0;
	 ret->data=(byte *)malloc(cap);
	 return ret;
}
void freebuf(buffer* buf) {
	 free(buf->data);
	 free(buf);
}

#if BYTE_ORDER==BIG_ENDIAN

inline unsigned short sus(unsigned short v) { return (v>>8)|(v<<8); }
inline short sss(short v) { return (signed)sus((unsigned)v); }
inline unsigned long sul(unsigned long v) { return (v>>24)|((v>>8)&0x0000ff00)|(0x00ff0000&(v<<8))|(v<<24); }
inline long ssl(long v) { return (signed)sul((unsigned)v); }

#else
#define sus(v) v
#define sss(v) v
#define sul(v) v
#define ssl(v) v
void endianconv(char* dat,const char* pat) {
}
#endif

void buf_wshort(buffer* buf, short val) {
	 checkbuffer(buf,buf->len+2);
	 buf->data[buf->len++]=val;
	 buf->data[buf->len++]=val>>8;
}
void buf_wlong(buffer* buf, long val) {
	 checkbuffer(buf,buf->len+4);
	 buf->data[buf->len++]=val;
	 buf->data[buf->len++]=val>>8;
	 buf->data[buf->len++]=val>>16;
	 buf->data[buf->len++]=val>>24;
}
void buf_wstr(buffer* buf, const char* str, int plen) {
	 int len=strlen(str);
	 if (!plen) plen=len+1;
	 int l2c;
	 if (len>plen) {
	l2c=plen-1;
	 } else {
	l2c=len+1;
	 }
	 checkbuffer(buf,buf->len+plen);
	 memcpy(buf->data+buf->len,str,l2c);
	 if (l2c<plen)
	memset(buf->data+buf->len+l2c,0,plen-l2c);
	 buf->len+=plen;
}
void buf_wbuf(buffer* buf, const void* dat, int len) {
	 if (buf->len+len>buf->cap) {
	resizebuffer(buf,buf->len+len);
	 }
	 memcpy(buf->data+buf->len,dat,len);
	 buf->len+=len;
}

buffer* buildit(it_info* it) {
	 int i,j,k;


#define wb(v) do { checkbuffer(buf,buf->len+1); buf->data[buf->len++]=v; }while (0)
#define ws(v) buf_wshort(buf,v);
#define wl(v) buf_wlong(buf,v);
#define wstr(v,l) buf_wstr(buf,v,l);
#define wsf(f) ws(it->f)
#define wlf(f) wl(it->f)
#define skip(l) checkbuffer(buf,buf->len+l); memset(buf->data+buf->len,0,l); buf->len+=l;
#define savel(v,val) *((unsigned long*)(buf->data+v))=sul(val);
#define saves(v,val) *((unsigned short*)(buf->data+v))=sus(val);

	 int msgofs=0;
	 int sampoffs=0;
	 int instoffs=0;
	 int patoffs=0;

	 int sampdatoffs=0;
	 buffer* buf=allocbuf(2048);

	 wl(IT_HEADSIG);
	 wstr(it->songname,26);
	 ws(0x1004);
	 wsf(ordnum);
	 wsf(insnum);
	 wsf(smpnum);
	 wsf(patnum);
	 ws(0x214);
	 ws(0x214);
	 // flags
	 ws(0x002d);
	 if (it->message) {
	ws(0x7);
	 } else {
	ws(0x6);
	 }
	 wb(it->globvol);
	 wb(0x2d);
	 wb(it->ispeed);
	 wb(it->itempo);
	 wb(it->pansep);
	 wb(0x0);
	 if (it->message) {
	int l=strlen(it->message);
	ws(l);
	msgofs=buf->len;
	skip(4);
	 } else {
	ws(0);
	wl(0);
	 }
	 skip(4);
	 for (i=0;i<64;i++)
	wb(it->pantable[i]);
	 for (i=0;i<64;i++)
	wb(it->voltable[i]);
	 
	 buf_wbuf(buf,it->order,it->ordnum);

	 instoffs=buf->len;
	 skip(it->insnum*4);

	 sampoffs=buf->len;
	 skip(it->smpnum*4);

	 patoffs=buf->len;
	 skip(it->patnum*4);

	 if (it->message) {
	savel(msgofs,buf->len)
	buf_wbuf(buf,it->message,strlen(it->message));
	 }
	 for (i=0;i<it->smpnum;i++) {
	it_sampinfo* smp=it->samples[i];
	savel(sampoffs+(i*4),buf->len);
	wl(IT_SAMPSIG);
	wstr(smp->dosfn,12);
	wb(0);
	wb(smp->gvol);
	wb(smp->flags);
	wb(smp->dvol);
	wstr(smp->name,26);
	wb(1);
	wb(smp->defaultpan);
	wl(smp->len);
	wl(smp->loopbeg);
	wl(smp->loopend);
	wl(smp->speed);
	wl(smp->sloopbeg);
	wl(smp->sloopend);
	sampdatoffs=buf->len;
	skip(4);
	wb(smp->vibspeed);
	wb(smp->vibdepth);
	wb(smp->vibrate);
	wb(smp->vibtype);

	savel(sampdatoffs,buf->len);
	int len=smp->len;
	if (smp->flags & SF_16BIT) 
		 len *= 2;
	buf_wbuf(buf,smp->sampdata,len);
	 }

	 for (i=0;i<it->insnum;i++) {
	it_instinfo* ins=it->insts[i];
	savel(instoffs+(i*4),buf->len);
	wl(IT_INSTSIG);
	wstr(ins->dosfn,12);
	wb(0);
	wb(ins->nna);
	wb(ins->dct);
	wb(ins->dca);
	ws(ins->fadeout);
	wb(ins->pps);
	wb(ins->ppc);
	wb(ins->gvol);
	wb(ins->dpan);
	wb(0);
	wb(0);
	ws(0);
	ws(0);
	wstr(ins->name,26);
	skip(6);
	checkbuffer(buf,buf->len+240);
	for (j=0;j<120;j++) {
		 buf->data[buf->len++]=ins->map_note[j];
		 buf->data[buf->len++]=ins->map_samp[j];
	}
	it_envelope* env=ins->env;
	for (j=0;j<3;j++,env++) {
		 wb(env->flags);
		 wb(env->nnodes);
		 wb(env->loopbeg);
		 wb(env->loopend);
		 wb(env->sloopbeg);
		 wb(env->sloopend);
		 for (k=0;k<25;k++) {
		wb(env->nodeval[k]);
		ws(env->nodepos[k]);
		 }
		 wb(0);
	}
	skip(7);
	 }
	 for (i=0;i<it->patnum;i++) {
	it_pattern* pat=it->pats[i];
	savel(patoffs+(i*4),buf->len);

	int lenoffs=buf->len;
	
	skip(2);
	ws(pat->rows);
	skip(4);
	unsigned char lastmask[64];
	unsigned char lastnote[64];
	unsigned char lastinst[64];
	unsigned char lastvolpan[64];
	unsigned char lasteff[64];
	unsigned char lasteffval[64];
	memset(lastmask,0,64);
	memset(lastnote,0,64);
	memset(lastinst,0,64);
	memset(lastvolpan,0,64);
	memset(lasteff,0,64);
	memset(lasteffval,0,64);

	int pstart=buf->len;
	note_t* cn=pat->notes;
	for (k=0;k<pat->rows;k++) {
		 for (j=0;j<pat->chans;j++,cn++) {
		int mv=0;
		if (cn->note) {
			 if (cn->note==lastnote[j]) mv |= 16;
			 else mv |= 1;
			 lastnote[j]=cn->note;
		} 
		if (cn->inst) {
			 if (cn->inst==lastinst[j]) mv |= 32;
			 else mv |= 2;
			 lastinst[j]=cn->inst;
		}
		if (cn->volpan) {
			 if (cn->volpan==lastvolpan[j]) mv |= 64;
			 else mv |= 4;
			 lastvolpan[j]=cn->volpan;
		}
		if (cn->eff || cn->effval) {
			 if (cn->eff==lasteff[j] && cn->effval==lasteffval[j]) mv |= 128;
			 else mv |= 8;
			 lastinst[j]=cn->inst;
		}
		if (mv) {
			 if (mv==lastmask[j]) {
			wb(j+1);
			 } else {
			wb((j+1)|128);
			wb(mv);
			lastmask[j]=mv;
			 }
			 if (mv & 1) wb(cn->note);
			 if (mv & 2) wb(cn->inst);
			 if (mv & 4) wb(cn->volpan);
			 if (mv & 8) { wb(cn->eff); wb(cn->effval); }
		}
		 }
		 wb(0);
	}
	int pend=buf->len;
	saves(lenoffs,pend-pstart);
	 }

	 return buf;
}
	  




#define SwapNone(X)
#if BYTE_ORDER==BIG_ENDIAN
#define Swap16(X) X=((X>>8)&0xFF)|((X<<8)&0xFF00)
#define Swap32(X) X=((X>>24)&0xFF)|((X>>8)&0xFF00)|((X<<8)&0xFF0000)|((X<<24)&0xFF000000)
#else
#define Swap16(X) 
#define Swap32(X) 
#endif

#define READER(type,name,size,swap)       \
type read_##name(FILE* f) {         \
	 type ret;               \
	 fread(&ret,size,1,f);         \
	 swap(ret);               \
	 return ret;               \
}

READER(unsigned short,u16,2,Swap16);
READER(unsigned int,u32,4,Swap32);

#define LOADER(type,name,func)         \
type name(const char* f) {         \
	 FILE* ff=fopen(f,"rb");         \
	 if (!ff) return NULL;         \
	 type ret=func(ff);            \
	 fclose(ff);               \
	 return ret;               \
}



kdmwaveset_t* read_waveset(FILE* f) {
	 kdmwaveset_t* ret=NEW(kdmwaveset_t);
	 read_u32(f);
	 ret->nwaves=read_u32(f);
	 ret->waves=NEWARR(kdmwave_t,ret->nwaves);
	 int i;
	 for (i=0;i<ret->nwaves;i++){
	kdmwave_t* cur=&ret->waves[i];
	cur->name=(char *)malloc(17);
	cur->name[16]=0;
	fread(cur->name,16,1,f);
	cur->len=read_u32(f);
	cur->repstart=read_u32(f);
	cur->repleng=read_u32(f);
	cur->finetune=read_u32(f);
	cur->snd=(byte *)malloc(cur->len+1);
	cur->repend=cur->repleng+cur->repstart;
	if (!cur->repleng) {
		 cur->repend=cur->len-8;
	}
	 }
	 for (i=0;i<ret->nwaves;i++){
	kdmwave_t* cur=&ret->waves[i];
	if (cur->len)
		 fread(cur->snd,cur->len,1,f);
	if (cur->repleng) {
		 cur->snd[cur->repstart+cur->repleng]=cur->snd[cur->repstart];
	} else {
		 cur->snd[cur->len]=0;
	}
	 }
	 return ret;
}

LOADER(kdmwaveset_t*,load_waveset,read_waveset);

#define RATEMUL 8259.417761282706762135
#define C5SPEED(finetune) ((int) (RATEMUL*((finetune)+748)/748.0))


it_sampinfo* convertsamp(kdmwave_t* in) {
	 it_sampinfo* ret=NEW(it_sampinfo);
	 memset(ret,0,sizeof(*ret));
	 ret->flags=1;
	 if (in->repleng) {
	ret->flags |= SF_LOOP;
	ret->loopbeg=in->repstart;
	ret->loopend=in->repend;
	 }
	 strncpy(ret->name,in->name,26);
	 ret->gvol=64;
	 ret->dvol=64;
	 ret->len=in->len;
	 ret->speed=C5SPEED(in->finetune);
	 ret->sampdata=(unsigned char *)malloc(in->len);
	 memcpy(ret->sampdata,in->snd,in->len);
	 int i;
	 for (i=0;i<in->len;i++)
	ret->sampdata[i] -= 128;
	 return ret;
}

it_instinfo* createinst(const char* name, int samp, int nna, it_envelope** env) {
	 it_instinfo* ret=NEW(it_instinfo);
	 memset(ret,0,sizeof(*ret));
	 ret->gvol=128;
	 ret->ppc=60;
	 int i;
	 for (i=0;i<120;i++) {
	ret->map_note[i]=i;
	ret->map_samp[i]=samp;
	 }
	 ret->nna=nna;

	 strncpy(ret->name,name,26);
	 
	 ret->dpan=128; // don't use default pan
	 if (env && env[0]) {
	memcpy(&ret->env[0],env[0],sizeof(it_envelope));
	 } else {
	ret->env[0].nnodes=2;
	ret->env[0].nodepos[0]=0;
	ret->env[0].nodeval[0]=64;
	ret->env[0].nodepos[1]=100;
	ret->env[0].nodeval[1]=64;
	 }
	 
	 if (env && env[1]) {
	memcpy(&ret->env[1],env[1],sizeof(it_envelope));
	 } else {
	ret->env[1].nnodes=2;
	ret->env[1].nodepos[0]=0;
	ret->env[1].nodeval[0]=0;
	ret->env[1].nodepos[1]=100;
	ret->env[1].nodeval[1]=0;
	 }
	 if (env && env[2]) {
	memcpy(&ret->env[2],env[2],sizeof(it_envelope));
	 } else {
	ret->env[2].nnodes=2;
	ret->env[2].nodepos[0]=0;
	ret->env[2].nodeval[0]=0;
	ret->env[2].nodepos[1]=100;
	ret->env[2].nodeval[1]=0;
	 }
	 return ret;
}

static inline void fillnote(note_t* not, int note, int inst, int volpan, int eff, int effval) {
	 not->note=note;
	 not->inst=inst;
	 not->volpan=volpan;
	 not->eff=eff;
	 not->effval=effval;
}

static inline void setnote(it_pattern* pat, int row, int chan, int note, int inst, int volpan, int eff, int effval) {
	 fillnote(&NOTEAT(pat,row,chan),note,inst,volpan,eff,effval);
}

typedef void (effectbuilder)(note_t* not, int row, int speed);

void pitch_effect_0(note_t* not, int row, int speed) {
	 if (row==0) {
	not->effval=0x56;
	 }
	 if (row*speed<256) {
	not->eff=8;
	 }
}
void pitch_effect_4(note_t* not, int row, int speed) {
	 if (row==0) {
	not->effval=0x46;
	 }
	 if (row*speed<256) {
	not->eff=8;
	 }
}
void vol_effect_0(note_t* not, int row, int speed) {
	 if (row==0) {
	not->effval=0x56;
	 }
	 if (row*speed<256) {
	not->eff='R'-'@';
	 }
}
void vol_effect_4(note_t* not, int row, int speed) {
	 if (row==0) {
	not->effval=0x4F;
	 }
	 if (row*speed<256) {
	not->eff='R'-'@';
	 }
}
void vol_effect_5(note_t* not, int row, int speed) {
	 if (row==0) {
	not->effval=0xFA;
	 }
	 if (row*speed<256) {
	not->eff='R'-'@';
	 }
}
void vol_effect_6(note_t* not, int row, int speed) {
	 
	 if (row==0) {
	not->effval=0x10;
	not->volpan=1;
	 }
	 if (row*speed<256) {
	not->eff=4;
	 }
}
void vol_effect_7(note_t* not, int row, int speed) {
	 
	 if (row==0) {
	not->effval=0x01;
	 }
	 if (row*speed<256) {
	not->eff=4;
	 }
}
effectbuilder* pitchefftbl[] = {
	 pitch_effect_0,
	 NULL,
	 NULL,
	 NULL,
	 NULL,
	 NULL,
	 NULL,
	 pitch_effect_4,
	 NULL,
	 NULL,
	 NULL,
	 NULL,
	 NULL,
};
effectbuilder* volefftbl[] = {
	 vol_effect_0,
	 NULL,
	 NULL,
	 NULL,
	 vol_effect_4,
	 vol_effect_5,
	 vol_effect_6,
	 vol_effect_7,
	 NULL,
};

it_envelope* pitchenv[9];
it_envelope* volenv[9];

it_info* convertkdm(const char* name, kdmfile_t* kdm, kdmwaveset_t* wav) {
	 int i,j,k;

	 it_info* inf=NEW(it_info);
	 memset(inf,0,sizeof(*inf));
	 strcpy(inf->songname,name);
	 for (i=0;i<64;i++) {
	inf->pantable[i]=32;
	inf->voltable[i]=64;

	 }
	 int maxquant=1;
	 for (i=0;i<kdm->ntracks;i++) {
	if (!kdm->trinst[i]) continue;
	if (maxquant<kdm->trquant[i]) maxquant=kdm->trquant[i];
	 }

	 inf->globvol=128;
	 inf->itempo=160;
	 inf->ispeed=64/maxquant;
	 
	 inf->insnum=0;
	 inf->smpnum=0;
	 inf->patnum=0;


	 inf->insts=NEWARR(it_instinfo*,128);
	 inf->samples=NEWARR(it_sampinfo*,128);
	 inf->pats=NEWARR(it_pattern*,250);

	 kdmnote_t* cur;

	 unsigned char sampmap[100];
	 unsigned char instmap[100][9][9];


	 unsigned char* trackmode=(unsigned char *)malloc(kdm->ntracks);
	 memset (trackmode,0,kdm->ntracks);
	 
	 for (cur=kdm->notes,i=0;i<kdm->nnotes;i++,cur++) {
	if (cur->vol1) {
		 if (trackmode[cur->track]==0)
		trackmode[cur->track]=2;
	} else {
		 trackmode[cur->track]=1;
	}
	 }

	 unsigned char* curpoly=(unsigned char *)malloc(kdm->ntracks);
	 unsigned char* maxpoly=(unsigned char *)malloc(kdm->ntracks);

	 unsigned int* noteexp=(unsigned int *)malloc(kdm->ntracks*128*4);

	 memset (sampmap,0,sizeof(sampmap));
	 memset (instmap,0,sizeof(instmap));
	 memset (maxpoly,0,kdm->ntracks);
	 memset (curpoly,0,kdm->ntracks);
	 memset (noteexp,0,kdm->ntracks*128*4);
	 it_envelope* cenv[3];
	 int transtime=kdm->notes[0].time;
	 for (cur=kdm->notes,i=0;i<kdm->nnotes;i++,cur++) {
	int cur_row=(((cur->time-transtime)<<6)/120)/inf->ispeed;
	int ct=cur->track;
	int cnp=(ct<<7)+cur->freq;
	for (j=0;j<128;j++) {
		 if (noteexp[(ct<<7)+j] && (noteexp[(ct<<7)+j]<=cur_row)) {
		curpoly[ct]--;
		noteexp[(ct<<7)+j]=0;
		 }
	}
	if (cur->vol1) {
		 if (!noteexp[cnp]) {
		curpoly[ct]++;
		if (maxpoly[ct]<curpoly[ct]) maxpoly[ct]=curpoly[ct];
		 }
		 if (trackmode[ct]!=2) {
		if (wav->waves[kdm->trinst[ct]].repleng) {
			 noteexp[cnp]=~0;
		} else {
			 int row_len=(((wav->waves[kdm->trinst[ct]].len<<7)/11025)/inf->ispeed);
			 noteexp[cnp]=cur_row+row_len;
		}
		 } else {
		noteexp[cnp]=cur_row+1;
		 }
	} else {
		 if (noteexp[cnp]) {
		curpoly[ct]--;
		noteexp[cnp]=0;
		 }
	}
	 }
	 free(curpoly);
	 free(noteexp);

	 unsigned char* trackchan = (unsigned char *)malloc(kdm->ntracks);
	 for (i=0,j=0;i<kdm->ntracks;i++) {
	trackchan[i]=j;
	int tvol=(kdm->trvol1[i]+kdm->trvol2[i])>>2;
	int tpan=((kdm->trvol2[i]-kdm->trvol1[i])>>3)+32;
	for (k=0;k<maxpoly[i];k++) {
		 inf->pantable[j+k]=tpan;
		 inf->voltable[j+k]=tvol;
	}
	j+=maxpoly[i];
	printf("%d: %d+%d, ti=%d (%s) tm=%d\n",i,trackchan[i],maxpoly[i],
			 kdm->trinst[i],wav->waves[kdm->trinst[i]].name,trackmode[i]);

	 }
	 int nchans=j;
	 unsigned int* channote=(unsigned int *)malloc(nchans*4);
	 int* chanstart=(int *)malloc(nchans*4);
	 unsigned int* chanexp=(unsigned int *)malloc(nchans*4);
	 memset (channote,0,nchans*4);
	 memset (chanexp,0,nchans*4);
	 memset (chanstart,0xff,nchans*4);



	 int last_pat=-1;
	 int last_row=-0;
	 int cur_pat=0;
	 int cur_row=0;
	 int pat_row=0;
	 it_pattern* pat=NULL;
	 effectbuilder** chaneffect=(effectbuilder **)malloc(nchans*sizeof(void*));
	 memset (chaneffect,0,nchans*sizeof(void*));
	 
	 for (cur=kdm->notes,i=0;i<kdm->nnotes;i++,cur++) {
	cur_row=(((cur->time-transtime)<<6)/120)/inf->ispeed;
	if (cur_row<0) cur_row=0;
	cur_pat=cur_row>>6;
	pat_row=cur_row&63;
	while (cur_pat != last_pat) {
		 pat=NEW(it_pattern);
		 pat->rows=64; pat->chans=nchans;
		 pat->notes=NEWARR(note_t,pat->rows*pat->chans);
		 memset(pat->notes,0,pat->rows*pat->chans*sizeof(note_t));

		 inf->pats[inf->patnum++]=pat;
		 last_pat++;
	}
	while (last_row<cur_row) {
		 for (j=0;j<nchans;j++) {
		if (chaneffect[j]) {
			 chaneffect[j](&NOTEAT(inf->pats[last_row>>6],last_row&63,j),last_row-chanstart[j],inf->ispeed);
		}
		 }
		 last_row++;
	}
	int ct=cur->track;

	int voleff=cur->voleff;
	int frqeff=cur->frqeff;
	if (voleff>8) voleff=0;
	if (frqeff>8) frqeff=0;
	int kinst=kdm->trinst[ct];
	int evoleff=voleff;
	int efrqeff=frqeff;
	if (!efrqeff || !pitchenv[efrqeff-1]) efrqeff=0;
	if (!evoleff || !volenv[evoleff-1]) evoleff=0;
	if (!instmap[kinst][evoleff][efrqeff]) {
		 if (!sampmap[kinst]) {
		inf->samples[inf->smpnum++]=convertsamp(&wav->waves[kinst]);
		sampmap[kinst]=inf->smpnum;
		 }
		 cenv[0]=evoleff?volenv[evoleff-1]:NULL;
		 cenv[1]=NULL;
		 cenv[2]=efrqeff?pitchenv[efrqeff-1]:NULL;
		 inf->insts[inf->insnum++]=
		createinst(wav->waves[kinst].name,sampmap[kinst],
				trackmode[ct]==2?1:0,cenv);
		 instmap[kinst][evoleff][efrqeff]=inf->insnum;
	}
	int inst=instmap[kinst][evoleff][efrqeff];
	if (cur->vol1) {
		 int chan=-1;
		 // check channel expirations
		 for (k=0;k<maxpoly[ct];k++) {
		int cc=trackchan[ct]+k;
		if (chanexp[cc] && chanexp[cc]<=cur_row) {
			 chanexp[cc]=0;
			 chanstart[cc]=-1;
			 chaneffect[cc]=NULL;
			 break;
		}
		 }
		 // check for channel playing this note
		 for (k=0;k<maxpoly[ct];k++) {
		int cc=trackchan[ct]+k;
		if (channote[cc]==cur->freq) {
			 chan=cc;
			 break;
		}
		 }
		 if (chan==-1) {
		chan=trackchan[ct];
		for (k=0;k<maxpoly[ct];k++) {
			 int cc=trackchan[ct]+k;
			 if (chanstart[cc]<chanstart[chan]) {
			chan=cc;
			 }
		}
		 }
		 chanstart[chan]=cur_row;
		 if (trackmode[ct]!=2) {
		if (wav->waves[kdm->trinst[ct]].repleng) {
			 chanexp[chan]=~0;
		} else {
			 int row_len=(((wav->waves[kdm->trinst[ct]].len<<7)/11025)/inf->ispeed);
			 chanexp[chan]=cur_row+row_len;
		}
		 } else {
		chanexp[chan]=cur_row+1;
		 }
		 channote[chan]=cur->freq;
		 if (frqeff && pitchefftbl[frqeff-1]) {
		chaneffect[chan]=pitchefftbl[frqeff-1];
		
		 } else if (voleff && volefftbl[voleff-1]) {
		chaneffect[chan]=volefftbl[voleff-1];
		 }
		 setnote(pat,pat_row,chan,cur->freq+35,inst,0,0,0);
	} else {
		 int chan=-1;
		 for (k=0;k<maxpoly[ct];k++) {
		int cc=trackchan[ct]+k;
		if (channote[cc]==cur->freq) {
			 chan=cc;
			 break;
		}
		 }
		 if (chan!=-1) {
		chanstart[chan]=-1;
		chanexp[chan]=0;
		channote[chan]=0;
		chaneffect[chan]=NULL;
		setnote(pat,pat_row,chan,254,0,0,0,0);
		 }
	}
	 }
	 if (!pat_row) {
	inf->patnum--;
	free(inf->pats[inf->patnum]);
	 }
	 cur_row--;

	 NOTEAT(inf->pats[cur_row>>6],cur_row&63,0).eff=3;

	 pat=inf->pats[0];
	 for (i=0;i<nchans;i++) {
	note_t* nt=&NOTEAT(pat,0,i);
	if (!nt->note) nt->note=254;
	 }

	 inf->ordnum=inf->patnum+1;
	 inf->order=(unsigned char *)malloc(inf->patnum+1);
	 for (i=0;i<inf->patnum;i++) {
	inf->order[i]=i;
	 }
	 inf->order[inf->patnum]=255;

	 free(trackchan);
	 free(chaneffect);
	 free(chanexp);
	 free(chanstart);
	 free(channote);
	 free(maxpoly);
	 free(trackmode);
	 return inf;

}

void buildenvelopes() {
	 int i;
	 for (i=0;i<9;i++) {
	volenv[i]=pitchenv[i]=NULL;
	 }
	 it_envelope* tmpa=NEW(it_envelope);
	 memset(tmpa,0,sizeof(it_envelope));
	 tmpa->flags = EF_ON;
	 tmpa->nnodes=2;
	 tmpa->nodepos[0]=0;
	 tmpa->nodepos[1]=15;
	 tmpa->nodeval[0]=2;
	 tmpa->nodeval[1]=0;

	 it_envelope* tmpb=NEW(it_envelope);
	 memcpy(tmpb,tmpa,sizeof(it_envelope));
	 tmpb->nodeval[0]=-4;

	 pitchenv[2]=tmpa;
	 pitchenv[1]=tmpb;

}
#define compar(f) if((a->f)<(b->f)) { return -1; } else if((a->f)>(b->f)) { return 1; }
#define compari(f) if((b->f)<(a->f)) { return -1; } else if((b->f)>(a->f)) { return 1; }
int compare_kdmnote (const kdmnote_t * a, const kdmnote_t * b)
{
	 compar (time);
	 compar (track);
	 compar (vol1?1:0);
	 compari (freq);
	 return a-b;
}

kdmfile_t* readkdm(FILE* f) {
	 kdmfile_t* ret=NEW(kdmfile_t);
	 kdmnote_t* cur;
	 int i;

	 read_u32(f);
	 ret->nnotes=read_u32(f);
	 ret->ntracks=read_u32(f);
	 byte* tmp2;
	 byte* tmp = (byte *)malloc(ret->ntracks*4);
	 fread(tmp,ret->ntracks*4,1,f);

#define QQ(f) ret->f=tmp; tmp += ret->ntracks
	 QQ(trinst);
	 QQ(trquant);
	 QQ(trvol1);
	 QQ(trvol2);
#undef QQ

	 tmp2 = tmp = (byte *)malloc(ret->nnotes*11);
	 ret->notes=(kdmnote_t*)malloc(ret->nnotes*sizeof(kdmnote_t));
	 fread(tmp,ret->nnotes*11,1,f);


	 for (cur=ret->notes,i=0;i<ret->nnotes;i++,tmp += 4)
	(cur++)->time=sul(*((unsigned int*)tmp));

#define QQ(f) for (cur=ret->notes,i=0;i<ret->nnotes;i++) (cur++)->f=*tmp++;
	 QQ(track);
	 QQ(freq);
	 QQ(vol1);
	 QQ(vol2);
	 QQ(frqeff);
	 QQ(voleff);
	 QQ(paneff);
#undef QQ
	 free(tmp2);

	 // Remove dead notes
	 kdmnote_t** noteplay=NEWARR(kdmnote_t*,128*ret->ntracks);
	 memset(noteplay,0,sizeof(kdmnote_t*)*128*ret->ntracks);
	 cur=ret->notes;
	 for (i=0;i<ret->nnotes;i++,cur++) {
	kdmnote_t** np=&noteplay[(cur->track<<7)+cur->freq];
	if (cur->vol1) {
		 *np=cur;
	} else {
		 if (*np && (*np)->time == cur->time) {
		(*np)->time=(unsigned)-1;
		cur->time=(unsigned)-1;
		 }
		 *np=NULL;
	}
	 }
	 free(noteplay);
	 qsort(ret->notes,ret->nnotes,sizeof(kdmnote_t),(int(*)(const void*,const void*))compare_kdmnote);
	 while (ret->nnotes && ret->notes[ret->nnotes-1].time == (unsigned)-1) ret->nnotes--;
	 return ret;
}
LOADER(kdmfile_t*,load_kdm,readkdm);

void usage(const char* pname) {
	 printf("Usage: %s [OPTION] INPUT [OUTPUT]\n",pname);
	 printf("Convert a Ken's Digital Music file to an Impulse Tracker Module\n");
	 printf("\n");
	 printf("  -w FILE       use waveset FILE instead of \"waves.kwv\"\n");
	 printf("  -s            don't create output, just list notes from file (for debugging)\n");
	 printf("\n");


}
#ifdef WIN32
#define PATH_SEP '\\'
#else
#define PATH_SEP '/'
#endif
int main(int argc, char** argv) {
	 char* pname=strrchr(argv[0],PATH_SEP);
	 if (pname) {
	pname++;
	 } else {
	pname=argv[0];
	 }

	 char* wavname="waves.kwv";
	 buildenvelopes();
	 argv++; argc--;
	 char* inf=NULL, *outf=NULL;
	 int mode=0;
	 while (argc--) {
	if (**argv=='-') {
		 if (!strcmp(*argv,"-s")) {
		mode=1;
		 } else if (!strcmp(*argv,"-w")) {
		if (!argc) {
			 printf("Error: option -w requires an argument.\n\n");
			 usage(pname);
			 exit(-1);
		}
		argc--;
		wavname = *++argv;
			 
		 } else {
		printf("Error: invalid option: %s\n\n",*argv);
		usage(pname);
		exit(-1);
		 }
	} else {
		 if (inf) {
		if (outf) {
			 usage(pname);
			 exit(-1);
		}
		outf=*argv;
		 } else {
		inf=*argv;
		 }
	}
	argv++;
	 }
	 if (!inf) {
	printf ("Malfunction. Need input.\n\n");
	usage(pname);
	exit(-1);
	 }
	 kdmwaveset_t* waves=load_waveset(wavname);
	 if(!(waves)){
	printf("Could not open %s: %s\n", wavname,strerror(errno));
	return 11;
	 }
	 kdmfile_t* kdmf=load_kdm(inf);
	 if(!(kdmf)){
	printf("Could not open %s: %s\n", inf,strerror(errno));
	return 1;
	 }
	 if (mode) {
	int i;
	kdmnote_t* cn=kdmf->notes;
	for (i=0;i<kdmf->nnotes;i++,cn++) {
		 printf("%8d %4d %4d %4d %4d %4d %4d %4d\n",cn->time,cn->freq,cn->track,cn->vol1,cn->vol2,
			cn->frqeff,cn->voleff,cn->paneff);
	}
	exit(0);
	 }
	 if (!outf) {
	outf=(char *)malloc(strlen(inf)+4);
	strcpy(outf,inf);
	char* dot=strrchr(outf,'.');
	char* s1=strrchr(outf,PATH_SEP);
	if (dot && dot>s1) {
		 strcpy(dot,".it");
	}
	 }
	 char* name=strrchr(inf,PATH_SEP);
	 if (name) {
	name++;
	 } else {
	name=inf;
	 }
	 it_info* it=convertkdm(name,kdmf,waves);

	 buffer* buf=buildit(it);
	 
	 FILE* f=fopen(outf,"wb");
	 fwrite(buf->data,1,buf->len,f);
	 fclose(f);
	 return 0;
}
